(*
    Copyright (C) 2012  G. Gonthier, B. Ziliani, A. Nanevski, D. Dreyer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

Require Import ssreflect ssrfun ssrbool ssrnat seq eqtype.
Require Import heaps.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


(*************************************************************************)
(* indom2 :                                                              *)
(*    similar to indom, but with the resolution of overlapping instances *) 
(* at the level of the operator.                                         *)
(*************************************************************************)

Definition invariant x (h : heap) :=
  def h -> x \in dom h.

(* Main structure and instances *)
Structure find (x : ptr) := 
  Form { heap_of :> heap; 
         _ : invariant x heap_of }.

Lemma found_pf A x (v : A) : invariant x (x :-> v).
Proof. by rewrite /invariant defPt domPt inE /= eq_refl. Qed.

Canonical ptr_found A x (v : A) :=
  @Form x (x :-> v) (@found_pf A x v).


(* Tagging for controling the instance search of :+ *)
Structure tagged_ptrU (x : ptr) := TagU { untagU :> ptr }.

Definition right_tagU (x : ptr) := TagU x x.
Canonical left_tagU x := right_tagU x.

Structure findU (h : heap) (x : ptr) :=
  FormU { ptr_ofU :> tagged_ptrU x;
          _ : invariant ptr_ofU h }.

Program Canonical search_leftU  x (h : heap) (f : find x) :=
  @FormU (f :+ h) x (left_tagU x) _.
Next Obligation.
case:f=>[i]; rewrite /invariant /= => H D.
by rewrite domUn !inE /= D (H (defUnl D)).
Qed.

Program Canonical search_rightU x (h : heap) (f : find x) :=
  @FormU (h :+ f) x (right_tagU x) _.
Next Obligation.
case: f=>[i]; rewrite /invariant /= => H D.
by rewrite domUn !inE /= D (H (defUnr D)) orbT.
Qed.


Structure equals_heap (h : heap) := 
  PackH { unpackH :> heap;
          _ : unpackH = h }.
Canonical equateH h := PackH (erefl h). 

Structure checkH (h1 h2 : heap) x :=
  CheckH { heap_ofH :> equals_heap h2;
           _ : invariant x (h1 :+ h2) }.

Program Canonical startH x h1 h2 (f : findU (h1 :+ h2) x) :=
  @CheckH h1 h2 f (equateH h2) _.
Next Obligation.
by case: f.
Qed.

Program Canonical findU_bridge x (h1 h2 : heap) (f : checkH h1 h2 x) :=
  @Form x (h1 :+ f) _.
Next Obligation.
by case: f=>[[_ /= ->] H].
Qed.

(* Main lemma *)
Lemma indom2 (x : ptr) (f : find x) : def f -> x \in dom f.
Proof. by case: f=>[i]; apply. Qed.


(*************************************************)
(*                   Examples                    *)
(*************************************************)

Example ex0 (x1 : ptr) :
          def (x1 :-> 1 :+ empty) -> 
          x1 \in dom (x1 :-> 1 :+ empty).
Proof.
move=>D.
by rewrite indom2.
Qed.

(* simple example *)
Example ex1 A (x1 x2 : ptr) (v1 v2 : A) (h1 h2 : heap) :
          def (h1 :+ x1 :-> 1 :+ (x2 :-> 3 :+ empty)) -> 
          if x2 \in dom (h1 :+ x1 :-> 1 :+ (x2 :-> 3 :+ empty)) 
            then 1 == 1 
            else 1 == 0.
Proof.
move=>D.
by rewrite indom2.
Qed.

(* same example, automatically unfolding a definition *)
Example ex2 A (x1 x2 : ptr) (v1 v2 : A) (h1 h2 : heap) :
          def (h1 :+ x1 :-> 1 :+ (x2 :-> 3 :+ empty)) -> 
          if x2 \in dom (h1 :+ x1 :-> 1 :+ (x2 :-> 3 :+ empty)) 
            then 1 == 1 
            else 1 == 0.
Proof.
set H := _ :+ _ :+ (_ :+ _).
move=>D.
by rewrite indom2.
Qed.

Example ex3 (x1 : ptr) :
          def ((empty :+ x1 :-> 1) :+ empty) -> 
          x1 \in dom ((empty :+ x1 :-> 1) :+ empty).
Proof.
move=>D.
by rewrite indom2.
Qed.


Example ex4 (x1 x2 x3 x4 : ptr) :
  let h := ((empty :+ (x1 :-> 1 :+ x3 :-> 3)) :+ (x4 :-> 4 :+ x2 :-> 2)) in
  def h -> 
  (x1 \in dom h) && (x2 \in dom h) && (x3 \in dom h) && (x4 \in dom h).
Proof.
move=>h D.
by rewrite !indom2.
Qed.