Require Import mtac.
Require Import ssreflect ssrfun ssrbool ssrnat seq eqtype.
Add LoadPath "../../examples-lo/".
Require Import heaps.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


(* Collect pointers in a heap *)
Module Scan.

(* The algorithm is defined as follows:
   - if the heap is h1 :+ h2, then recurse over h1 and h2 and concatenate the
     results.
   - if the heap is x :-> v, then return [x]
   - otherwise, return []
*)

Definition axiom h s :=
  def h -> uniq s /\ forall x, x \in s -> x \in dom h.

(* Main structure *)
Record form h := Form {seq_of :> seq ptr; axiom_of : axiom h seq_of}.

Lemma union_pf h1 h2 (s1 : form h1) (s2 : form h2) : 
        axiom (h1 :+ h2) (s1 ++ s2).
Proof.
move:s1 s2=>[i1 H1] [i2 H2]; rewrite /axiom /= in H1 H2 * => D.
case/(_ (defUnl D)): H1=>U1 H1; case/(_ (defUnr D)): H2=>U2 H2.
split=>[|x]; last first. 
- rewrite mem_cat; case/orP; [move/H1 | move/H2]; 
  by rewrite domUn !inE /= D => -> //=; rewrite orbT.
rewrite cat_uniq U1 U2 andbT -all_predC.
apply/allP=>x; move/H2=>H3; apply: (introN idP); move/H1=>H4.
by case: defUn D=>// _ _; move/(_ _ H4); rewrite H3.
Qed.

Lemma ptr_pf A x (v : A) : axiom (x :-> v) [:: x].
Proof. 
rewrite /axiom /= defPt => D; split=>//.
by move=>y; rewrite inE; move/eqP=>->; rewrite domPt inE /= eq_refl D.
Qed.

Lemma default_pf h : axiom h [::].
Proof. by move=>D; split. Qed.

Definition scan_un h1 h2 (f1 : form h1) (f2 : form h2) := Form (union_pf f1 f2).
Definition scan_p x A (v : A) := Form (@ptr_pf _ x v).
Definition scan_h h := Form (@default_pf h).

(** Mtactic scan for collecting the pointers of a heap *)
Definition scan :=
  mfix f (h : heap) : M (form h) :=
    mmatch h with
    | [x A (v:A)] x :-> v => ret (scan_p x v)
    | [l r] l :+ r => 
      rl <- f l;
      rr <- f r;
      ret (scan_un rl rr)
    | [h] h => ret (scan_h h)
    end.

End Scan.


(* Search a pointer in a list. Could be generalize to any type *)
Module Search.

(* The algorithm is defined as follow:
   - test if the list is (x :: s) for x being the element we are looking for
   - if the list is (y :: s), then recurse using s
*)
Section Search.
Variable A : eqType.

Lemma in_head (x : A) s : x \in (x :: s).
Proof. by rewrite inE eq_refl. Qed.

Lemma in_cons (x y : A) s (f : x \in s) : x \in (y :: s).
Proof. by move:f; rewrite /= inE orbC=>->. Qed.

Set Printing Existential Instances.

Definition NotFound : Exception.  exact exception. Qed.

Program Definition search x :=
  mfix f (s : seq A) : M (x \in s) :=
    mmatch s as s' return M (x \in s') with 
    | [s'] x :: s' => ret (in_head x s')
    | [y s'] y :: s' => 
      r <- f s';
      ret (in_cons y r)
    | _ => raise NotFound
    end.

End Search.
End Search. 


(* Search for two different pointers in a list *)
Module Search2.

(* The algorithm works as follow: Let x and y be the pointers we are looking for
   - If we found x, then search for y using the previous module
   - If we found y, then search for x using the previous module
   - If, instead, we found some pointer z, then recurse
*)

Definition axiom (x y : ptr) (s : seq ptr) := 
  uniq s -> x != y.

Lemma foundx_pf x y s (f : y \in s) : axiom x y (x :: s).
Proof.
case/andP=>H2 _. 
case: eqP; last by [].  
move=>H; rewrite H in H2.
by rewrite f in H2.
Qed.

Lemma foundy_pf x y s (f : x \in s) : axiom x y (y :: s).
Proof.
case/andP=>H2 _. 
case: eqP; last by [].  
move=>H; rewrite H in f.
by rewrite f in H2.
Qed.

Lemma foundz_pf x y z s (f : axiom x y s) : axiom x y (z :: s).
Proof.
move: f.
rewrite /axiom /= =>H3. 
by case/andP=>_; apply: H3.
Qed.

Definition NotFound : Exception.  exact exception. Qed.

Definition search2 x y :=
  mfix f (s : _) : M (axiom x y s) :=
    mmatch s with
    | [s'] x :: s' =>
      r <- Search.search y s';
      ret (foundx_pf r)
    | [s'] y :: s' =>
      r <- Search.search x s';
      ret (foundy_pf r)
    | [z s'] z :: s' =>
      r <- f s';
      ret (foundz_pf r)
    | _ => raise NotFound
    end. 

End Search2.


(* Now package everything together *)
Module NoAlias.


Lemma combine h x y (sc : Scan.form h) 
  (sch2 : Search2.axiom x y (Scan.seq_of sc)) :
  def h ->  x != y.
Proof.
move=> D.
apply: sch2. apply: (proj1 (Scan.axiom_of sc D)).
Qed.


Definition noalias h (D : def h) : M (forall x y, M (x != y)) :=
  ps <- Scan.scan h;
  ret (fun x y => 
    s2 <- Search2.search2 x y (Scan.seq_of ps);
    retS (combine s2 D)).

End NoAlias.

Notation noalias := NoAlias.noalias.

Program
Definition interactive_search2 x y :=
  mfix f (s : _) : M (Search2.axiom x y s) :=
    mmatch s as s' return M (uniq s' -> x != y) with
    | [s'] x :: s' =>
      r <- Search.search y s';
      ret _
    | [s'] y :: s' =>
      r <- Search.search x s';
      ret _
    | [z s'] z :: s' =>
      r <- f s';
      ret _ 
    | _ => raise Search2.NotFound
    end. 
Next Obligation.
  move=>x y _ _ _ s' _ f /=.
  case/andP. 
  case: eqP; last by [].  
  by move=>->; rewrite f.
Qed.
Next Obligation.
  by intros; apply: (Search2.foundy_pf r).
Qed.
Next Obligation.
  by intros; apply: (Search2.foundz_pf (z:=z) r).
Qed.


(** Examples *)

Example ex0 A (x1 x2 x3 : ptr) (v1 v2 v3 : A) :
  def (x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3) ->
 [/\ x1 != x2, x2 != x3 & x1 != x3].
Proof.
move=>D.
pose H := run (noalias D).
by rewrite !(eval (H _ _)).
Qed.

Example exnc A (x1 x2 x3 x4 : ptr) (v1 v2 : A) (h1 h2 : heap) :
  def (h1 :+ x2 :-> 1 :+ h2 :+ x1 :-> v2 :+ (x3 :-> v1 :+ empty)) -> 
     (x1 != x2) /\
     (x1 != x2) && (x2 != x3) && (x3 != x1) /\ 
     (x2 == x3) = false /\ (x1 == x2) = false /\
     ((x1 != x2) && (x2 != x3)) = (x1 != x2) /\
     ((x1 != x2) && (x2 != x3)) = (x1 != x2) /\
     ((x1 != x2) && (x2 != x3)) = (x1 != x2) /\
     (x1 != x2) && (x2 != x3) && ((x1 != x4) || (x3 != x1)).
Proof.
move=>D.
pose H := run (noalias D).
split.
  by apply: (run (H _ _)). 
split.
  (* backwards reasoning works *)
  by rewrite !(eval (H _ _)). 
split.
  (* subterm selection works *)
  by rewrite [x2 == x3](negbTE (eval (H _ _))).
split.
  (* composition works *)
  by rewrite (negbTE (eval (H _ _))).
split.
  by rewrite [x2 != x3](eval (H _ _)) andbT.
split.
  by rewrite (eval (H x2 _)) andbT.
split.
  by rewrite (eval (H _ x3)) andbT.
(* rewriting skips the subgoals that don't apply *)
(* just as it should *)
rewrite !(negbTE (eval (H _ _))).
by rewrite orbT.
Abort.


Example ex_stress A (x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 : ptr)
              (x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 : ptr)
              (x21 x22 x23 x24 x25 x26 x27 x28 x29 x30 : ptr)
              (v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 : A) 
              (h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 : heap) :
          def (h1 :+ h2 :+ h3 :+ h4 :+ h5 :+ 
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               x16 :-> v6 :+ x17 :-> v7 :+ x18 :-> v8 :+ x19 :-> v9 :+ x20 :-> v10 :+ 
               x11 :-> v1 :+ x12 :-> v2 :+ x13 :-> v3 :+ x14 :-> v4 :+ x15 :-> v5 :+ 
               x6 :-> v6 :+ x7 :-> v7 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               x21 :-> v1 :+ x22 :-> v2 :+ x23 :-> v3 :+ x24 :-> v4 :+ x25 :-> v5 :+ 
               x26 :-> v6 :+ x27 :-> v7 :+ x28 :-> v8 :+ x29 :-> v9 :+ x30 :-> v10 :+ 
               h6 :+ h7 :+ h8 :+ h9 :+ h10) -> 
          (x1 != x2) && (x2 != x3) && (x4 != x5) && (x5 != x6) && (x7 != x8) && (x9 != x10) &&
          (x11 != x12) && (x12 != x13) && (x14 != x15) && (x15 != x16) && (x17 != x18) && (x19 != x20).
Proof.
move=>D.
Time pose H := run (noalias D);
apply/andP;split;
do ![apply/andP;split|];
apply: (run (H _ _)). (* 12s *)
Time Qed. (* 2s *)


Example ex_stress_rw A (x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 : ptr)
              (x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 : ptr)
              (x21 x22 x23 x24 x25 x26 x27 x28 x29 x30 : ptr)
              (v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 : A) 
              (h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 : heap) :
          def (h1 :+ h2 :+ h3 :+ h4 :+ h5 :+ 
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               x16 :-> v6 :+ x17 :-> v7 :+ x18 :-> v8 :+ x19 :-> v9 :+ x20 :-> v10 :+ 
               x11 :-> v1 :+ x12 :-> v2 :+ x13 :-> v3 :+ x14 :-> v4 :+ x15 :-> v5 :+ 
               x6 :-> v6 :+ x7 :-> v7 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               x21 :-> v1 :+ x22 :-> v2 :+ x23 :-> v3 :+ x24 :-> v4 :+ x25 :-> v5 :+ 
               x26 :-> v6 :+ x27 :-> v7 :+ x28 :-> v8 :+ x29 :-> v9 :+ x30 :-> v10 :+ 
               h6 :+ h7 :+ h8 :+ h9 :+ h10) -> 
          (x1 != x2) && (x2 != x3) && (x4 != x5) && (x5 != x6) && (x7 != x8) && (x9 != x10) &&
          (x11 != x12) && (x12 != x13) && (x14 != x15) && (x15 != x16) && (x17 != x18) && (x19 != x20).
Proof.
move=>D.
Time pose H := run (noalias D);
by rewrite !(eval (H _ _)). (* 10s *)
Time Qed. (* 1s *)


Require Import noalias.

Example ex_stress' A (x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 : ptr)
              (x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 : ptr)
              (x21 x22 x23 x24 x25 x26 x27 x28 x29 x30 : ptr)
              (v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 : A) 
              (h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 : heap) :
          def (h1 :+ h2 :+ h3 :+ h4 :+ h5 :+ 
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               x16 :-> v6 :+ x17 :-> v7 :+ x18 :-> v8 :+ x19 :-> v9 :+ x20 :-> v10 :+ 
               x11 :-> v1 :+ x12 :-> v2 :+ x13 :-> v3 :+ x14 :-> v4 :+ x15 :-> v5 :+ 
               x6 :-> v6 :+ x7 :-> v7 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               x21 :-> v1 :+ x22 :-> v2 :+ x23 :-> v3 :+ x24 :-> v4 :+ x25 :-> v5 :+ 
               x26 :-> v6 :+ x27 :-> v7 :+ x28 :-> v8 :+ x29 :-> v9 :+ x30 :-> v10 :+ 
               h6 :+ h7 :+ h8 :+ h9 :+ h10) -> 
          (x1 != x2) && (x2 != x3) && (x4 != x5) && (x5 != x6) && (x7 != x8) && (x9 != x10) &&
          (x11 != x12) && (x12 != x13) && (x14 != x15) && (x15 != x16) && (x17 != x18) && (x19 != x20).
Proof.
move=>D.
Time pose H := noaliasR D;
apply/andP;split;
do ![apply/andP;split|];
apply: (H _ _). (* 15s *)
Time Qed. (* 2s *)


Example ex_stress'_rw A (x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 : ptr)
              (x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 : ptr)
              (x21 x22 x23 x24 x25 x26 x27 x28 x29 x30 : ptr)
              (v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 : A) 
              (h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 : heap) :
          def (h1 :+ h2 :+ h3 :+ h4 :+ h5 :+ 
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               x16 :-> v6 :+ x17 :-> v7 :+ x18 :-> v8 :+ x19 :-> v9 :+ x20 :-> v10 :+ 
               x11 :-> v1 :+ x12 :-> v2 :+ x13 :-> v3 :+ x14 :-> v4 :+ x15 :-> v5 :+ 
               x6 :-> v6 :+ x7 :-> v7 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               x21 :-> v1 :+ x22 :-> v2 :+ x23 :-> v3 :+ x24 :-> v4 :+ x25 :-> v5 :+ 
               x26 :-> v6 :+ x27 :-> v7 :+ x28 :-> v8 :+ x29 :-> v9 :+ x30 :-> v10 :+ 
               h6 :+ h7 :+ h8 :+ h9 :+ h10) -> 
          (x1 != x2) && (x2 != x3) && (x4 != x5) && (x5 != x6) && (x7 != x8) && (x9 != x10) &&
          (x11 != x12) && (x12 != x13) && (x14 != x15) && (x15 != x16) && (x17 != x18) && (x19 != x20).
Proof.
move=>D.
Time pose H := noaliasR D;
by rewrite !(H _ _). (* 8s *)
Time Qed. (* 3s *)
