Add LoadPath "../examples-lo".
Add LoadPath "./htt".
Require Import ssreflect ssrbool ssrnat ssrfun eqtype seq rels.
Require Import prelude prefix xfindM mtac.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(*********************************************)
(* Syntactic cancellation of nat expressions *)
(*********************************************)

(* addition of option-nats *)

Definition oplus (o1 o2 : option nat) := 
  match o1, o2 with 
    Some n1, Some n2 => Some (n1 + n2)
  | _, _ => None
  end. 

Notation "o1 \+ o2" := (oplus o1 o2) (at level 43, left associativity).

Definition def A (o : option A) := if o is Some _ then true else false. 

Lemma oaddnC o1 o2 : o1 \+ o2 = o2 \+ o1.
Proof. by case: o1 o2=>[n1|][n2|] //=; rewrite addnC. Qed.

Lemma oaddnA o1 o2 o3 : o1 \+ (o2 \+ o3) = o1 \+ o2 \+ o3. 
Proof. by case: o1 o2 o3=>[n1|][n2|][n3|] //=; rewrite addnA. Qed.

Lemma oaddn0 o : oplus o (Some 0) = o. 
Proof. by case: o=>[n|] //=; rewrite addn0. Qed.

Lemma oadd0n o : oplus (Some 0) o = o.
Proof. by rewrite oaddnC oaddn0. Qed.

Lemma oaddnCA o1 o2 o3 : o1 \+ (o2 \+ o3) = o2 \+ (o1 \+ o3).
Proof. by rewrite oaddnC (oaddnC o1) oaddnA. Qed. 

Lemma oaddnAC o1 o2 o3 : o1 \+ o2 \+ o3 = o1 \+ o3 \+ o2.
Proof. by rewrite (oaddnC o1) -oaddnA oaddnC. Qed. 

Lemma oaddnK o o1 o2 : def (o \+ o1) -> o \+ o1 = o \+ o2 -> o1 = o2. 
Proof. 
case: o o1 o2=>[n|][n1|][n2|] //= _ []; case/eqP.
by rewrite eqn_add2l; move/eqP=>->. 
Qed.

Lemma oaddSome n1 n2 : Some n1 \+ Some n2 = Some (n1 + n2).
Proof. by []. Qed.

Definition oleq (o1 o2 : option nat) := 
  match o1, o2 with 
    Some n1, Some n2 => n1 <= n2
  | _, _ => false
  end. 

Notation "o1 \<= o2" := (oleq o1 o2) (at level 41).  

Lemma oleq_refl o : def o -> o \<= o.
Proof. by case: o=>[n|] /=. Qed.

Lemma oleq_trans o1 o2 o3 : o1 \<= o2 -> o2 \<= o3 -> o1 \<= o3. 
Proof. by case: o1 o2 o3=>[n1|][n2|][n3|] //; apply: leq_trans. Qed.

Lemma def_add o1 o2 : def (o1 \+ o2) = (def o1) && (def o2).
Proof. by case: o1 o2=>[n1|][n2|]. Qed.

Lemma def_addl o1 o2 : def (o1 \+ o2) -> def o1. 
Proof. by case: o1. Qed.

Lemma def_addr o1 o2 : def (o1 \+ o2) -> def o2. 
Proof. by rewrite oaddnC; apply: def_addl. Qed.

Lemma leq_defl o1 o2 : o1 \<= o2 -> def o1.
Proof. by case: o1 o2=>[n1|][n2|]. Qed.

Lemma leq_defr o1 o2 : o1 \<= o2 -> def o2.
Proof. by case: o1 o2=>[n1|][n2|]. Qed.

Lemma leqKl o o1 o2 : def (o \+ o1) -> (o \+ o1) \<= (o \+ o2) = o1 \<= o2.
Proof. by case: o o1 o2=>[n|][n1|][n2|] //= _; rewrite leq_add2l. Qed.

Lemma leqKr o o1 o2 : def (o1 \+ o) -> (o1 \+ o) \<= (o2 \+ o) = o1 \<= o2.
Proof. by move=>D; rewrite -!(oaddnC o) in D *; apply: leqKl. Qed.

Lemma leq_inl o1 o2 : def (o1 \+ o2) -> o1 \<= (o1 \+ o2).
Proof. by case: o1 o2=>[n1|][n2|] //= _; apply: leq_addr. Qed.

Lemma leq_inr o1 o2 : def (o1 \+ o2) -> o2 \<= (o1 \+ o2).
Proof. by rewrite oaddnC; apply: leq_inl. Qed.

(* variable contexts *)

Inductive ctx := Context {nat_ctx : seq nat}.
Definition empc := Context [::].

(* syntactic nats *)
(* nat variables are syntactified as indices in the context *)
(* plus is syntactified as concatenation *)
(* succ is syntactified as adding 1 to the constant term *)
(* zero is ignored *)

Inductive var := Var of nat.
Definition term := prod (seq var) nat. (* the second nat is the constant term *)

(* lookup functions for vars in a context *)
Definition vlook := [fun i => onth (nat_ctx i)].

(* interpretation function for a single variable *)
Definition einterp G v := let: Var i := v in vlook G i.

(* interpretation function for sequences of vars and for terms *)

Fixpoint interp_vars G s :=
  if s is v :: s' then 
    if s' is ([::])%core then einterp G v else einterp G v \+ interp_vars G s'
 else Some 0.

Notation interp G t := (interp_vars G t.1 \+ Some t.2). 

Lemma interpvars_cons G v s : 
        interp_vars G (v :: s) = einterp G v \+ interp_vars G s.
Proof. by case: s=>//=; rewrite oaddn0. Qed.

Lemma interpvars_cat i s1 s2 : 
        interp_vars i (s1 ++ s2) = interp_vars i s1 \+ interp_vars i s2.
Proof.
elim: s1 s2=>[|v s1 IH] s2; first by rewrite oadd0n.
by rewrite cat_cons !interpvars_cons IH oaddnA.
Qed.

Inductive fact_vars := LeqVars of seq var & seq var.
Inductive fact := Leq of term & term.

(* returns the proposition generated by a fact *)
Definition exec_vars G fv c1 c2 :=
  let: LeqVars s1 s2 := fv in 
    (interp_vars G s1 \+ Some c1) \<= (interp_vars G s2 \+ Some c2).

Definition exec G f := 
  let: Leq t1 t2 := f in exec_vars G (LeqVars t1.1 t2.1) t1.2 t2.2. 

(* functions to collect nat variable indices out of a term *)

Fixpoint vars s := if s is Var i :: s' then i :: vars s' else [::]. 

(* function to remove one occurence of a variable from a sequence *)

Fixpoint remove_var n s :=
  if s is Var m :: s' then 
    if m == n return seq var then s' else Var m :: remove_var n s'
  else [::].

Lemma remove_def G i s : 
        def (interp_vars G s) -> def (interp_vars G (remove_var i s)).
Proof. 
elim: s=>// [[k]] s IH.
rewrite interpvars_cons def_add; case/andP=>D1 D2 /=.
case: eqP=>// _; case E: (remove_var _ _)=>[|v s'] //.
by rewrite -E interpvars_cons def_add D1 (IH D2).
Qed.

Lemma einterp_def G i s : 
        def (interp_vars G s) -> i \in vars s -> def (einterp G (Var i)).
Proof.
elim: s=>// [[k]] s IH.
rewrite interpvars_cons => D /=; rewrite inE.
case/orP; last by apply: IH (def_addr D).
by move/eqP=>->; apply: def_addl D.
Qed.

Lemma remove_leq G i s : 
        def (interp_vars G s) -> 
        interp_vars G (remove_var i s) \<= interp_vars G s.
Proof.
elim: s=>// [[k]] s IH; rewrite interpvars_cons /= => D.
case: ifP=>E; first by apply: leq_inr.
rewrite interpvars_cons leqKl; first by apply: IH (def_addr D). 
by rewrite !def_add in D *; case/andP: D=>->; apply: remove_def. 
Qed.

Lemma remove_eq G i s :   
        def (interp_vars G s) -> i \in vars s -> 
        interp_vars G s = einterp G (Var i) \+ interp_vars G (remove_var i s). 
Proof.
elim: s i G=>// [[k]] s IH i G; rewrite interpvars_cons => D /=.
rewrite inE; case/orP; first by move/eqP=>->{i}; rewrite eq_refl.
move=>N; case: ifP=>E; first by move/eqP: E=>?; subst k. 
rewrite interpvars_cons oaddnCA; congr (_ \+ _).
by apply: IH (def_addr D) N.
Qed.

(* main function: simplifies a leq expression by cancelling *)
(* common terms on both sides. if the left side becomes empty *)
(* we can decide that the initial left expression was less-then-or-equal *)
(* to the initial right expression *)

Fixpoint cancel_vars (G : ctx) (s1 s2 : seq var) : bool := 
  if s1 is Var n :: s1' return bool then
    if n \in vars s2 return bool then cancel_vars G s1' (remove_var n s2)
    else false 
  else true. 
        
(**********************)
(* Decision procedure *)
(**********************)

(* We know left <= right if the cancelation algorithm matches every
variable in left to something in right (thus, empying left side), and
the constant factor in left is smaller than the constant factor in
right. *)

Definition decision G t1 t2 := 
  if cancel_vars G t1.1 t2.1 return bool then t1.2 <= t2.2 else false. 

Lemma decisionvar_sound G s1 s2 c1 c2 : 
        def (interp G (s1, c1)) ->
        def (interp G (s2, c2)) ->
        c1 <= c2 ->
        cancel_vars G s1 s2 ->
        interp G (s1, c1) \<= interp G (s2, c2).
Proof.
elim: s1 s2=>[|[i] s1 IH] s2 D1 D2 H. 
- move=>_; apply: oleq_trans (leq_inr D2)=>/=.
  by rewrite add0n. 
rewrite !interpvars_cons /= -oaddnA in D1 *.
case: ifP=>E //.
move/(IH _ (def_addr D1) _ H).
rewrite def_add (remove_def _ (def_addl D2)) (def_addr D2). 
rewrite (remove_eq (def_addl D2) E) -!oaddnA.
by move/(_ (erefl _)); rewrite leqKl.
Qed.

Lemma decision_sound G t1 t2 : 
        def (interp G t1) -> def (interp G t2) ->
        decision G t1 t2 -> interp G t1 \<= interp G t2.
Proof.
rewrite /decision /cancel; case: t1 t2=>s1 c1 [s2 c2].
move=>D1 D2; case: ifP=>//= H E. 
by apply: decisionvar_sound D1 D2 E H. 
Qed.

(**************************)
(* Reflection into syntax *)
(**************************)

(* first a bunch of lemmas on context extension *)

Definition subctx G D := prefix (nat_ctx G) (nat_ctx D). 

Lemma subctx_refl G: subctx G G.
Proof. by []. Qed.

Hint Resolve subctx_refl.

Lemma subctx_trans G D E : subctx G D -> subctx D E -> subctx G E.
Proof. by apply: prefix_trans. Qed.

Lemma def_sub_einterp G D v : 
        subctx G D -> def (einterp G v) -> def (einterp D v).
Proof. by case: v=>n P /=; case E: (onth _ _)=>[k|] // _; move/P: E=>->. Qed.

Lemma def_sub_interpvars G D s : 
        subctx G D -> def (interp_vars G s) -> def (interp_vars D s). 
Proof.
elim: s=>// [[k]] s IH X; rewrite !interpvars_cons !def_add. 
by case/andP; move/(def_sub_einterp X)=>->; move/(IH X).
Qed.

Lemma def_sub_interp G D t : 
        subctx G D -> def (interp G t) -> def (interp D t).
Proof.
move=>X; rewrite !def_add; case/andP.
by move/(def_sub_interpvars X)=>->->.
Qed.

Lemma sub_interpvars G D s : 
        def (interp_vars G s) -> subctx G D -> 
        interp_vars G s = interp_vars D s.
Proof.
elim: s=>// [[k]] s IH; rewrite !interpvars_cons def_add.
case/andP=>/= D1 D2 X; case E: (onth _ _) D1=>[n|] // _.
by rewrite (X _ _ E) (IH D2 X). 
Qed.

Lemma sub_interp G D t : 
        def (interp G t) -> subctx G D -> interp G t = interp D t.
Proof.
case: t=>s c; rewrite def_add /= andbT => D'.
by move/(sub_interpvars D')=>->. 
Qed.

(* now the ast structure which is what the *)
(* reflection procedure returns *)

Definition invariant G D t n := [/\ interp D t = Some n & subctx G D].

(* inputs: G - input contex, n - nat to be reflected
outputs: term_of - reflected nat, ctx_of - output context (extending G) *)

Structure ast (G : ctx) (n : nat) := 
  Ast {term_of : term; 
       ctx_of : ctx;
       _ : invariant G ctx_of term_of n}.

(* the invariant ensures that the reflected term_of *)
(* is a reflection of the input expression *)

Definition syn_zero : term := ([::], 0).
Definition syn_succ (t : term) : term := (t.1, t.2.+1).
Definition syn_add (t1 t2 : term) : term := (t1.1 ++ t2.1, t1.2 + t2.2).
Definition syn_var (i : nat) : term := ([:: Var i], 0).

Lemma zero_pf G : invariant G G syn_zero 0.
Proof. by []. Qed.

Lemma succ_pf G n (r : ast G n) : 
        invariant G (ctx_of r) (syn_succ (term_of r)) n.+1.
Proof.
case: r=>[[s c]] D /= [H X]; split=>//=.
by rewrite -addn1 -oaddSome oaddnA H /= addn1.
Qed.

Lemma add_pf G n1 n2 (r1 : ast G n1) (r2 : ast (ctx_of r1) n2) : 
        invariant G (ctx_of r2) (syn_add (term_of r1) (term_of r2)) (n1 + n2).
Proof.
case: r1 r2=>[[s1 c1]] D1 /= [H1 X1][[s2 c2] D2 [H2 X2]] /=. 
have [DD1 DD2] : def (interp D1 (s1, c1)) /\ def (interp D2 (s2, c2)).
- by rewrite H1 H2.
split; last by apply: subctx_trans X1 X2.
rewrite interpvars_cat /= -oaddSome -!oaddnA -(oaddnCA (Some c1)).
by rewrite oaddnA -(sub_interp DD1 X2) H1 H2.
Qed.

Lemma var_pf G v (r : xfind v (nat_ctx G)) : 
        invariant G (Context (seq_of r)) (syn_var (index_of r)) v.
Proof. by case: r=>s i [H1 H2]; split=>//=; rewrite H1 /= addn0. Qed.

Definition to_ast : forall G n, M (ast G n) := 
  mfix f G n :=
  mmatch n with
  | 0 => retS (Ast (@zero_pf G))
  | [n'] n'.+1 =>
      r <- f G n';
      retS (Ast (@succ_pf G n' r))
  | [n1 n2] n1 + n2 =>
      r1 <- f G n1; r2 <- f (ctx_of r1) n2;
      retS (Ast (@add_pf G n1 n2 r1 r2))
  | [v] v =>
      r <- findx v (nat_ctx G);
      retS (Ast (@var_pf G v r))
  end. 

Implicit Arguments to_ast [].

(* Now let's try the decision procedure *)

Lemma decision_pf n1 n2 (r1 : ast empc n1) (r2 : ast (ctx_of r1) n2) :
        decision (ctx_of r2) (term_of r1) (term_of r2) -> n1 <= n2. 
Proof.
case: r1 r2=>t1 G1 [H1 X1][t2 G2 [H2 /= X2]].
have [DD1 DD2] : def (interp G1 t1) /\ def (interp G2 t2).
- by rewrite H1 H2.
move: (sub_interp DD1 X2)=>E; rewrite {}E in DD1 H1.
by move/(decision_sound DD1 DD2); rewrite H1 H2. 
Qed.

Definition NotLeqException : Exception. exact: exception. Qed.

Program Definition decisionR n1 n2 : M (n1 <= n2) := 
  r1 <- to_ast empc n1;
  r2 <- to_ast (ctx_of r1) n2;
  mmatch decP (@idP (decision (ctx_of r2) (term_of r1) (term_of r2))) return M (n1 <= n2) with
  | [pf] left pf => ret (decision_pf (r2:=r2) pf)
  | _ => raise NotLeqException : M (n1 <= n2)
  end. 

Definition ex1 := Eval hnf in run (decisionR 0 1).

Print ex1.

(* the output is not readable, but at least we didn't get an exception *)

Fail Check (run (decisionR 1 0)). 

(* Now we package this in the index-out-of-bounds check for sized
vectors. *)

Structure vector (n : nat) : Type := 
  Vector {seq_of : seq nat; _ : size seq_of = n}.

Definition vsize n (v : vector n) := n.

(* the accessor function *)

Lemma my_head_nil : 0 < @size nat [::] -> nat. 
Proof. by []. Qed.

Definition myhead s : 0 < size s -> nat := 
  if s is x :: _ then fun _ => x else my_head_nil.

Lemma behead_pf i (s : seq nat) : i.+1 < size s -> i < size (behead s).
Proof. by rewrite size_behead -subn1 ltn_subRL addnC addn1. Qed.

Fixpoint ith' s i : i < size s -> nat :=
  if i is i'.+1 then fun pf => ith' (behead_pf pf)
  else @myhead s.

Lemma ith_pf n (v : vector n) i : i < n -> i < size (seq_of v).
Proof. by case: v=>s /= ->. Qed.

Definition ith n (v : vector n) i (pf : i < n) := ith' (ith_pf v pf). 

(* The problem with the accessor function is that you always have to
pass the proof of i < n. But often times, we know what n and i
are. For example, in mathematics, n can often be fixed to a specific
dimension (say, 3-dimensional vectors, because we only work in 3d
space).  In that case, we can avoid providing these proofs by hand.
One option is to define the projections fst, snd, thrd, to which we
pass the proofs of 0 < 3, 1 < 3 and 2 < 3 directly.  But that has a
problem that if we fix n to 4, we need to redo the work for 0 < 4, 1 <
4, 2 < 4, and also add a new projection corresponding to 3 < 4.

Here's another approach which uniformly defines these projections for
any n, in one fell swoop. Of course, the projections typecheck only
when n and i are constants, but the work of defining them doesn't need
to be redone from scratch for every n. *)

Notation "[ 'ith' v i ]" := (@ith _ v i (run (decisionR i.+1 (vsize v))))
(at level 0, v, i at level 8, format "[ 'ith'  v  i ]").

(* Now we can use the notation naturally. This exhibits the usefulness
of automation appearing directly in definitions rather than only in
proofs, as it must be in VeriML, Beluga and Ltac. *)

Definition inner_prod (v1 v2 : vector 3) := 
             [ith v1 0] * [ith v2 0] + 
             [ith v1 1] * [ith v2 1] + 
             [ith v1 2] * [ith v2 2].

(* In fact, we are much more general, and can use variables in indexes. *)

Definition silly1 n k (v : vector (n + k + 1)) := [ith v k] + [ith v n].

(* Notice that if we tried to k.+1 or n.+1 in the right-hand side, it won't work,
because neither k.+1 nor n.+1 are always smaller than n + k + k + 1. *)

Fail Check (fun n k (v : vector (n + k + k + 1)) => [ith v k.+1] + [ith v n.+1]).

(* but if we add 1 to the vector dimension, then it works *)

Definition silly2 n k (v : vector (n + k + k + 1 + 1)) := [ith v k.+1] + [ith v n.+1].

